/*-
 * Copyright (C)2023..2025 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2023..2025 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: pgoblin5pgeo.c,v 1.26 2025/02/09 22:04:06 babolo Exp $"

#define BLIN_COMPAT      4
#define Bpars_COMPAT     4
#define MULAR_COMPAT     0
#define MIFE_COMPAT      5
#define PGOBLIN_COMPAT   5
#define PGOBLIN_INTERNAL 1
#define PGOBLIN_MODULE   pgoblin_module

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sysexits.h>
#include <limits.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <math.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <multilar.h>
#include <mife.h>
#include <pgoblin.h>
#include "pgoblin5pgeo.h"
#include "gtype.h"

static u_int8_t pgoblin_formlen[] = {3, 4, 4, 5, 5, 6, 5, 6, 1};
static u_int8_t pgoblin_formbyt[] = {2, 4, 4, 8, 6, 12, 8, 16, 1, 0, 0, 0, 0, 0, 0, 0};
static char const *names[] =
{ ""     , "point"   , "line", "lseg"
, "box"  , "circle"  , "path", "polygon"
, "path+", "polygon+"
};

static int
/*****************************************************************************************************
 *****************************************************************************************************
 **                                                                                                 **/
init(pgoblin_exenv *exenv, u_int slot) {                                                           /**
 **                                                                                                 **
 *****************************************************************************************************
 *****************************************************************************************************/
    (void)exenv;
    (void)slot;
    return(0);
}

static int
/*****************************************************************************************************
 *****************************************************************************************************
 **                                                                                                 **/
fini(pgoblin_exenv *exenv) {                                                                       /**
 **                                                                                                 **
 *****************************************************************************************************
 *****************************************************************************************************/
    (void)exenv;
    return(0);
}

static int
/*****************************************************************************************************
 **                                                                                                 **/
coloin(pgoblin_rst *rst, const char *colosrc, u_int64_t *colores) {                                /**
 **                                                                                                 **
 *****************************************************************************************************/
#   define blin_internal_flags (rst->style->flags & BLIN_MASK)
    ssize_t    clen = 0;
    BLIN_flag  fl   = 0;
    char       rev[17];
    char      *oo;

    ifBLIN_QX3("+ %08X %s", rst->style->flags, colosrc);
    if  ('#' == *colosrc) colosrc++;
    if  (16 < (clen = (ssize_t)strlen(colosrc))) {
        ifBLIN_QX0("Color descriptor too long, %d", (int)clen);
        clen = -EX_USAGE;
        errno = EDOM;
        goto out;
    }
    oo = rev;
    if  (1 < clen) for (ssize_t i = clen; i > 0; i -= 2) {
        *(oo++) = colosrc[i - 2];
        *(oo++) = colosrc[i - 1];
    }
    rev[clen] = '\0';
    if  (  (PGOBLIN_NOFORM != (PGOBLIN_NOFORM & rst->style->format))
        && (pgoblin_formbyt[rst->style->format] != clen)
        ) {
        ifBLIN_QX0( "Different format foreground and background colors %"BLIN_D"<>%d"
                  , clen
                  , pgoblin_formbyt[rst->style->format]
                  );
        clen = -EX_USAGE;
        errno = ENOTSUP;
        goto out;
    }
    switch(clen) {
    case 1:
        if  ('0' != (*colosrc & 0x00FE)) {
            ifBLIN_QX0("Illegal color for bitmap");
            clen = -EX_USAGE;
            errno = EDOM;
            goto out;
        }
        fl = PGOBLIN_BIT;
        break;
    case 2:
        break;
    case 4:
        if  (!!(PGOBLIN_LONG & rst->style->format)) {
            fl = PGOBLIN_LONG;
        } else {
            fl = PGOBLIN_ALFA;
        }
        break;
    case 6:
        fl = PGOBLIN_COLOR;
        break;
    case 8:
        if  (PGOBLIN_LONG & rst->style->format) {
            fl = PGOBLIN_LONG | PGOBLIN_ALFA;
        } else {
            fl = PGOBLIN_COLOR | PGOBLIN_ALFA;
        }
        break;
    case 12:
        fl = PGOBLIN_LONG | PGOBLIN_COLOR;
        break;
    case 16:
        fl = PGOBLIN_LONG | PGOBLIN_COLOR | PGOBLIN_ALFA;
        break;
    default:
        ifBLIN_QX0("Illegal length of color descriptor , %d", (int)clen);
        clen = -EX_DATAERR;
        errno = EDOM;
        goto out;
    }
    if  ((PGOBLIN_NOFORM != (PGOBLIN_NOFORM & rst->style->format)) && (rst->style->format != fl)) {
        ifBLIN_QX0("Color format not match: %04X <> %04X", rst->style->format, fl);
        clen = -EX_USAGE;
        errno = EINVAL;
        goto out;
    }
    rst->style->format = fl;
    errno = 0;
    *colores = strtoul(rev, &oo, 16);
    if  (errno) {
        ifBLIN_QW0("strtoul() color=%s~", colosrc);
        clen = -EX_USAGE;
        goto out;
    }
    if  (!!*oo) {
        ifBLIN_QX0("Illegal symbol in color=%s~", colosrc);
        clen = -EX_USAGE;
        errno = EINVAL;
        goto out;
    }
    if  (PGOBLIN_LONG & rst->style->format) {
        *colores = ((*colores >>  8) & 0x00FF00FF00FF00FFULL)
                 | ((*colores <<  8) & 0xFF00FF00FF00FF00ULL)
        ;
    }
out:
    ifBLIN_QX3("- %"BLIN_P"d %08X %016"BLIN_O"X", clen, rst->style->flags, *colores);
    return((int)clen);
#   undef blin_internal_flags
}


static int
/*****************************************************************************************************
 **                                                                                                 **/
flags(pgoblin_rst *rst, char **srcstyle) {                                                         /**
 **                                                                                                 **
 *****************************************************************************************************/
#   define blin_internal_flags (rst->flags & BLIN_MASK)
    int              length;
    const char      *Tflag  = NULL;
    const char      *flar   = "=:b:Bf:h:Lw:x:X:y:Y:";
    babolo_opts     *bos    = NULL;
    babolo_parm     *bp     = NULL;
    int              ex     = 0;          /*     < 0  */
    char            *oo;
    const char      *o;
    int              c;

    ifBLIN_QX3("+ =%s~", *srcstyle);
    bp = babolo_getparm(Bpars_NOAL | Bpars_LFLN | Bpars_NOEN | Bpars_FAST, srcstyle, "#", 0);
    if  (!bp) {
        ifBLIN_QW0("getparm");
        ex = -EX_SOFTWARE;
        goto out;
    }
    ex = (int)bp->cntr;
    ifBLIN_QX5("%d %"BLIN_X, bp->argc, BLIN_I(*srcstyle));
    if  (!(bos = babolo_openopts(Bpars_ESYM | 'E', 'Z'))) {
        ifBLIN_QW0("openopts");
        ex = -EX_SOFTWARE;
        goto out;
    }
    if  (!!babolo_setparm(bos, (blin_internal_flags >> 1) & BLIN_MASK, bp, flar)) {
        ifBLIN_QW0("setopts");
        ex = -EX_SOFTWARE;
        goto out;
    }
    for (length = 0; !!(c = babolo_getopts(bos));) {
        ifBLIN_QX5("-%c", c);
        switch (c) {
        case 'b': if  (!(o = babolo_getoptsarg(bos))) {
                ;     ifBLIN_QW0("No -%c arg", bos->c);
                ;     ex = -EX_USAGE;
                ;     goto out;
                ; }
                ; if  (0 > (length = coloin(rst, o, &rst->style->bg))) {
                ;     ifBLIN_QW0("Background color %s error", o);
                ;     ex = length;
                ;     goto out;
                ; }
                ; break;
        case 'B': if  (8 > rst->style->format) {
                ;     ifBLIN_QX0("Illegal -%c flag", bos->c);
                ;     ex = -EX_USAGE;
                ;     errno = ENOSPC;
                ;     goto out;
                ; }
                ; rst->style->format |= PGOBLIN_BIT;
                ; break;
        case 'f': if  (!(o = babolo_getoptsarg(bos))) {
                ;     ifBLIN_QW0("No -%c arg", bos->c);
                ;     ex = -EX_USAGE;
                ;     goto out;
                ; }
                ; if  (0 > (length = coloin(rst, o, &rst->style->fg0))) {
                ;     ifBLIN_QW0("Foreground color %s error", o);
                ;     ex = length;
                ;     goto out;
                ; }
                ; break;
        case 'h': if  (!(o = babolo_getoptsarg(bos))) {
                ;     ifBLIN_QW0("No -%c arg", bos->c);
                ;     ex = -EX_USAGE;
                ;     goto out;
                ; }
                ; oo = o;
                ; if  ('/' != *o) {
                ;     errno = 0;
                ;     rst->style->scaleyup = strtod(o, &oo);
                ;     if  (errno) {
                ;         ifBLIN_QW0("-%c strtod scaleyup", bos->c);
                ;         ex = -EX_USAGE;
                ;         goto out;
                ; }   }
                ; if  ('/' == *(oo++)) {
                ;     errno = 0;
                ;     rst->style->scaleydown = strtod(oo, NULL);
                ;     if  (errno) {
                ;         ifBLIN_QW0("-%c strtod scaleydown", bos->c);
                ;         ex = -EX_USAGE;
                ;         goto out;
                ; }   }
                ; break;
        case 'L': if  (PGOBLIN_LONG & rst->style->format) break;
                ; if  ((0 < length) && (12 > length)) {
                ;     ifBLIN_QX0("Sign LONG before colors");
                ;     ex = -EX_USAGE;
                ;     errno = ENOSPC;
                ;     goto out;
                ; }
                ; rst->style->format |= PGOBLIN_LONG;
                ; break;
        case 'w': if  (!(o = babolo_getoptsarg(bos))) {
                ;     ifBLIN_QW0("No -%c arg", bos->c);
                ;     ex = -EX_USAGE;
                ;     goto out;
                ; }
                ; oo = o;
                ; if  ('/' != *o) {
                ;     errno = 0;
                ;     rst->style->scalexup = strtod(o, &oo);
                ;     if  (errno) {
                ;         ifBLIN_QW0("-%c strtod scalexup", bos->c);
                ;         ex = -EX_USAGE;
                ;         goto out;
                ; }   }
                ; if  ('/' == *(oo++)) {
                ;     errno = 0;
                ;     rst->style->scalexdown = strtod(oo, NULL);
                ;     if  (errno) {
                ;         ifBLIN_QW0("-%c strtod scalexdown", bos->c);
                ;         ex = -EX_USAGE;
                ;         goto out;
                ; }   }
                ; break;
        case 'x': errno = 0;
                ; rst->style->marginxdown = (int)babolo_getoptsnum(bos, 0);
                ; if  (errno) {
                ;     ifBLIN_QW0("-%c babolo_getoptsnum marginxdown", bos->c);
                ;     ex = -EX_USAGE;
                ;     goto out;
                ; }
                ; break;
        case 'X': errno = 0;
                ; rst->style->marginxup = (int)babolo_getoptsnum(bos, 0);
                ; if  (errno) {
                ;     ifBLIN_QW0("-%c babolo_getoptsnum marginxup", bos->c);
                ;     ex = -EX_USAGE;
                ;     goto out;
                ; }
                ; break;
        case 'y': errno = 0;
                ; rst->style->marginydown = (int)babolo_getoptsnum(bos, 0);
                ; if  (errno) {
                ;     ifBLIN_QW0("-%c babolo_getoptsnum marginydown", bos->c);
                ;     ex = -EX_USAGE;
                ;     goto out;
                ; }
                ; break;
        case 'Y': errno = 0;
                ; rst->style->marginyup = (int)babolo_getoptsnum(bos, 0);
                ; if  (errno) {
                ;     ifBLIN_QW0("-%c babolo_getoptsnum marginyup", bos->c);
                ;     ex = -EX_USAGE;
                ;     goto out;
                ; }
                ; break;
        case '=': if  (!(Tflag = babolo_getoptsarg(bos))) exit(EX_USAGE);
                ; switch(Tflag[0]) {
                  case '0': blin_ctl(BLIN_CTL_DUMP);  exit(EX_OK);
                  case '1': babolo_optest(flar);      exit(EX_OK);
                  case '2': babolo_optext(bos);       exit(EX_OK);
                  case '3': babolo_dumpopts(bos);     exit(EX_OK);
                  default:                            break;
                ; }
        default : ifBLIN_QX0("Illegal flag -%c", bos->c);
                ; ex = -EX_USAGE;
                ; errno = EINVAL;
                ; goto out;
    }   }
out:
    if  (!!bos) babolo_closeopts(bos);
    bos = NULL;
    ifBLIN_QX3("- %d=%s~", ex, *srcstyle);
    return(ex);
#   undef blin_internal_flags
}

static void
/*****************************************************************************************************
 *****************************************************************************************************
 **                                                                                                 **/
dump(pgoblin_rst *rst, FILE *ouf) {                                                                /**
 **                                                                                                 **
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (rst->flags & BLIN_MASK)
    ifBLIN_QX3("+ [%c](%d)", pgoblin_regn[rst->nst], rst->nst);
    if  (!rst->style) {
        ifBLIN_QX0("Dump pgeo: no style @STY %c", pgoblin_regn[rst->nst]);
        errno = EINVAL;
        goto out;
    }
    fprintf( ouf
           , "+######################################%.*s\n"
           , PGOBLIN_STRING_ID_LEN
           , rst->style->id
           );
    fprintf( ouf
           , "%08X b#%016"BLIN_O"X  f#%016"BLIN_O"X x%f/%f y%f/%f\n"
             "margin=(%d..%d)x(%d..%d)y fill=(%d..%d)x(%d..%d)y\n"
           , rst->style->flags
           , rst->style->bg
           , rst->style->fg0
           , rst->style->scalexup
           , rst->style->scalexdown
           , rst->style->scaleyup
           , rst->style->scaleydown
           , rst->style->marginxdown
           , rst->style->marginxup
           , rst->style->marginydown
           , rst->style->marginyup
           , rst->style->fillxdown
           , rst->style->fillxup
           , rst->style->fillydown
           , rst->style->fillyup
           );
    for (u_int i = 0; i < PGOBLIN_GEO_MAXCOLOR + 1; ++i) {
        if  (0 > rst->style->sz[i]) continue;
        fprintf(ouf, "%s[%d]\n", names[i], rst->style->sz[i]);
        for (int k = 0; k <= rst->style->sz[i]; ++k) {
            if  (!!rst->style->fg[i][k].def) {
                fprintf( ouf
                       , "    %d[%d] #%016"BLIN_O"X\n"
                       , k
                       , rst->style->fg[i][k].nmsz
                       , rst->style->fg[i][k].fg
                       );
                for (int n = 0; n < rst->style->fg[i][k].nmsz; ++n) {
                    fprintf( ouf
                           , "        %d #%016"BLIN_O"X=%s~\n"
                           , n
                           , rst->style->fg[i][k].namec[n].fg
                           , rst->style->fg[i][k].namec[n].name
                           );
                }
            } else if (0 < rst->style->fg[i][k].nmsz) {
                fprintf( ouf
                       , "    %d[%d]\n"
                       , k
                       , rst->style->fg[i][k].nmsz
                       );
                for (int n = 0; n < rst->style->fg[i][k].nmsz; ++n) {
                    fprintf( ouf
                           , "        %d #%016"BLIN_O"X=%s~\n"
                           , n
                           , rst->style->fg[i][k].namec[n].fg
                           , rst->style->fg[i][k].namec[n].name
                           );
    }   }   }   }
    fprintf( ouf
           , "-######################################%.*s\n"
           , PGOBLIN_STRING_ID_LEN
           , rst->style->id
           );
    fflush(ouf);
out:
    ifBLIN_QX3("-");
    return;
#   undef blin_internal_flags
}

static int
/*****************************************************************************************************
 *****************************************************************************************************
 **                                                                                                 **/
parser(pgoblin_rst *rst, char *srcstyle) {                                                         /**
 **                                                                                                 **
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (rst->flags & BLIN_MASK)
    int           ex      = EX_OK;                    /*                                  */

    ifBLIN_QX3("+ %"BLIN_X, BLIN_I(rst));
    if  (!rst->style) {
        int           lfcount;                        /*                                 */
        confchain    *newcf   = NULL;
        confchain    *conf    = NULL;
        babolo_parm  *bp;
        int           sz;

        if  (!(rst->style = calloc(1, sizeof(pgoblin_realsty)))) {
            ifBLIN_QX0("No memory #1 for STY %c", pgoblin_regn[rst->nst]);
            ERROUT(-EX_OSERR, ENOMEM);
        }
        strncpy(rst->style->id, PGOBLIN_STYLE_ID, strlen(PGOBLIN_STYLE_ID));
        rst->style->flags = blin_internal_flags;
        rst->style->format = PGOBLIN_NOFORM;
        rst->style->wideline = 1;
        rst->style->bg = 0x0000000000000000ULL;
        rst->style->fg0 = 0xFFFFFFFFFFFFFFFFULL;
        rst->style->scalexup = 1.0;
        rst->style->scalexdown = 1.0;
        rst->style->scaleyup = 1.0;
        rst->style->scaleydown = 1.0;
        sz = 1 << (PGOBLIN_SZLEV1 + PGOBLIN_SZLEV2 + PGOBLIN_SZLEV3 - 1);
        rst->style->marginxup = rst->style->marginyup = sz - 1;
        rst->style->marginxdown = rst->style->marginydown = -sz;
        rst->style->fillxup = rst->style->fillyup = -sz;
        rst->style->fillxdown = rst->style->fillydown = sz - 1;
        if  (0 > (lfcount = flags(rst, &srcstyle))) {
            ifBLIN_QW0("parser flags @STY %c(%d)", pgoblin_regn[rst->nst], rst->nst);
            ex = lfcount;
            goto out;
        }
        for (u_int i = 0; i <= PGOBLIN_GEO_MAXCOLOR; ++i) rst->style->sz[i] = -1;
        for (; !!*srcstyle;) { /*    ,    confchain *conf */
            int type;

            bp = babolo_getparm( Bpars_LFLN | Bpars_FIXB | Bpars_NOEN | Bpars_NOAL | 5
                               , &srcstyle
                               , NULL
                               , 0
                               );
            if  (!bp) {
                ifBLIN_QW0("babolo_getparm line %d", lfcount);
                ex = -EX_USAGE;
                goto out;
            }
            if  (0 == bp->argc) {
                babolo_freeparm(bp);
                break;
            }
            if  (2 > bp->argc) {
                ifBLIN_QX0("babolo_getparm line %d short %d", lfcount, bp->argc);
                ex = -EX_USAGE;
                errno = EINVAL;
                goto out;
            }
            if  (4 < bp->argc) {
                ifBLIN_QX0("babolo_getparm line %d long %d", lfcount, bp->argc);
                ex = -EX_USAGE;
                errno = EINVAL;
                goto out;
            }
            lfcount += bp->cntr;
            type = 0;
            if  (!!*(bp->argv[0]) && (0 >= (type = (int)babolo_testword(&gtype, bp->argv[0])))) {
                ifBLIN_QW0("babolo_testword=%s~", bp->argv[0]);
                ex = -EX_SOFTWARE;
                goto out;
            }
            /* , ,   */
            if  (!(newcf = calloc(1, sizeof(confchain)))) {
                ifBLIN_QW0("No memory STY %c(%d) line %d", pgoblin_regn[rst->nst], rst->nst, lfcount);
                ex = -EX_OSERR;
                goto out;
            }
            /*     newcf */
            newcf->col = -1;
            newcf->type = (u_int8_t)type;
            for (int i = 1; i < bp->argc; ++i) {
                if  ('#' == *(bp->argv[i])) {
                    ex = coloin( rst
                               , bp->argv[i]
                               , &newcf->fg
                               );
                    if  (0 > ex) {
                        ifBLIN_QW0("coloin=%s~", bp->argv[1]);
                        goto out;
                    }
                    ex = EX_OK;
                    newcf->chn = (!conf) ? 1 : (conf->chn + 1);
                } else if (('0' <= *(bp->argv[i]) && ('9' >= *(bp->argv[i])))) {
                    int y;

                    errno = 0;
                    y = (int)strtoul(bp->argv[i], NULL, 10);
                    if  (!!errno) {
                        ifBLIN_QW0( "Illegal number STY %c(%d) line %d"
                                  , pgoblin_regn[rst->nst]
                                  , rst->nst
                                  , lfcount
                                  );
                        ex = -EX_DATAERR;
                        goto out;
                    }
                    if  (0 < y) {
                        newcf->col = y;
                        if  (rst->style->sz[type] < y) rst->style->sz[type] = y;
                    }
                } else { /* XXXX   " () */
                    newcf->name = bp->argv[i]; /* Bpars_NOAL   strdup() */
            }   }
            if  (!newcf->name) {
                ifBLIN_QX6( "%"BLIN_X" %016"BLIN_O"X col%d N%u type%u nx%"BLIN_X
                          , BLIN_I(newcf)
                          , newcf->fg
                          , newcf->col
                          , newcf->chn
                          , newcf->type
                          , BLIN_I(newcf->next)
                          );
            } else {
                ifBLIN_QX6( "%"BLIN_X" %016"BLIN_O"X col%d=%s~ N%u type%u nx%"BLIN_X
                          , BLIN_I(newcf)
                          , newcf->fg
                          , newcf->col
                          , newcf->name
                          , newcf->chn
                          , newcf->type
                          , BLIN_I(newcf->next)
                          );
            }
            if  (!newcf->chn) { /*     ,  */
                ifBLIN_QX1( "Void conf string STY %c(%d) line %d"
                          , pgoblin_regn[rst->nst]
                          , rst->nst
                          , lfcount
                          );
                free(newcf);
            } else { /*      */
                if  (0 > newcf->col) newcf->col = 0;
                if  (rst->style->sz[type] < 0) rst->style->sz[type] = 0;
                newcf->next = conf;
                conf = newcf;
            }
            newcf = NULL;
            babolo_freeparm(bp);
            if  (!conf->name) {
                ifBLIN_QX5( "%"BLIN_X" %016"BLIN_O"X col%d N%u type%u nx%"BLIN_X
                          , BLIN_I(conf)
                          , conf->fg
                          , conf->col
                          , conf->chn
                          , conf->type
                          , BLIN_I(conf->next)
                          );
            } else {
                ifBLIN_QX5( "%"BLIN_X" %016"BLIN_O"X col%d=%s~ N%u type%u nx%"BLIN_X
                          , BLIN_I(conf)
                          , conf->fg
                          , conf->col
                          , conf->name
                          , conf->chn
                          , conf->type
                          , BLIN_I(conf->next)
                          );
        }   }
        /*         */
        for (u_int i = 0; i <= PGOBLIN_GEO_MAXCOLOR; ++i) {
            ifBLIN_QX7("[%d]%d", i, rst->style->sz[i] + 1);
            if  (0 > rst->style->sz[i]) continue;
            if  (!(rst->style->fg[i] = calloc((size_t)rst->style->sz[i] + 1, sizeof(confar)))) {
                ifBLIN_QW0("No memory STY %c(%d) line %d", pgoblin_regn[rst->nst], rst->nst, lfcount);
                ex = -EX_OSERR;
                goto out;
        }   }
        for (; !!conf; conf = newcf) { /*  */
            newcf = conf->next;
            if  (!conf->name) { /*    ,   */
                rst->style->fg[conf->type][conf->col].fg = conf->fg;
                rst->style->fg[conf->type][conf->col].def = 1;
                free(conf);
            } else { /*    ,      */
                conf->next = rst->style->fg[conf->type][conf->col].names;
                rst->style->fg[conf->type][conf->col].names = conf;
        }   }
        for (u_int i = 0; i <= PGOBLIN_GEO_MAXCOLOR; ++i) {  /*      */
            for (int k = 0; k <= rst->style->sz[i]; ++k) { /*             */
                coname *nmlist;

                conf = rst->style->fg[i][k].names;
                for (rst->style->fg[i][k].nmsz = 0; !!conf; conf = conf->next) {
                    if  (!!conf->name) rst->style->fg[i][k].nmsz++;
                }
                if  (!rst->style->fg[i][k].nmsz) continue;
                if  (!(nmlist = malloc((size_t)rst->style->fg[i][k].nmsz * sizeof(coname)))) {
                    ifBLIN_QW0( "No memory STY %c(%d) line %d"
                              , pgoblin_regn[rst->nst]
                              , rst->nst
                              , lfcount
                              );
                    ex = -EX_OSERR;
                    goto out;
                }
                conf = rst->style->fg[i][k].names;
                for (int n = 0; n < rst->style->fg[i][k].nmsz; ++n) {
                    nmlist[n].name = conf->name;
                    nmlist[n].fg = conf->fg;
                    newcf = conf;
                    conf = conf->next;
                    free(newcf);
                }
                rst->style->fg[i][k].namec = nmlist;
        }   }
        if  (PGOBLIN_NOFORM == (PGOBLIN_NOFORM & rst->style->format)) {
            rst->style->format = PGOBLIN_COLOR | PGOBLIN_ALFA;
        }
    } else {
        if  (!!strncmp(rst->style->id, PGOBLIN_STYLE_ID, strlen(PGOBLIN_STYLE_ID))) {
            ifBLIN_QX0("Illegal style %s @STY %c", rst->style->id, pgoblin_regn[rst->nst]);
            ERROUT(-EX_DATAERR, EINVAL);
        }
        if  (8 < rst->style->format) {
            ifBLIN_QX0( "Illegal style format %d @STY %c(%u)"
                      , rst->style->format
                      , pgoblin_regn[rst->nst]
                      , rst->nst
                      );
            ERROUT(-EX_DATAERR, EINVAL);
    }   }
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static int
/*****************************************************************************************************
 *****************************************************************************************************
 **                                                                                                 **/
clear(pgoblin_rst *rst) {                                                                          /**
 **                                                                                                 **
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (rst->flags & BLIN_MASK)
    int    ex      = EX_OK; /*   */
    void **t;

    ifBLIN_QX3("+ %"BLIN_X, BLIN_I(rst));
    if  (!!rst) {
        if  (!!rst->style) {
            if  (!!strncmp(rst->style->id, PGOBLIN_STYLE_ID, strlen(PGOBLIN_STYLE_ID))) {
                ifBLIN_QX0("Illegal style %s @STY %c", rst->style->id, pgoblin_regn[rst->nst]);
                ERROUT(EX_DATAERR, EINVAL);
            }
            if  (8 < rst->style->format) {
                ifBLIN_QX0( "Illegal style format %d @STY %c(%u)"
                          , rst->style->format
                          , pgoblin_regn[rst->nst]
                          , rst->nst
                          );
                ERROUT(EX_DATAERR, EINVAL);
            }
            for (int i = 0; i < (1 << (PGOBLIN_SZLEV1 * 2)); ++i) {
                t = rst->style->grole[i];
                if  (!!t) {
                    for (int k = 0; k < (1 << (PGOBLIN_SZLEV2 * 2)); ++k) {
                        if  (!!t[k]) {
                            free(t[k]);
                            t[k] = NULL;
                    }   }
                    free(rst->style->grole[i]);
                    rst->style->grole[i] = NULL;
            }   }
            free(rst->style);
            rst->style = NULL;
    }   }
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static void
/*****************************************************************************************************
 **                                                                                                 **/
deco(int idx, int *i) {                                                                            /**
 **                                                                                                 **
 *****************************************************************************************************/
    int    m;

    idx += 1 << (PGOBLIN_SZLEV1 + PGOBLIN_SZLEV2 + PGOBLIN_SZLEV3 - 1);
    i[0] = idx >> (PGOBLIN_SZLEV2 + PGOBLIN_SZLEV3);
    m = (1 << PGOBLIN_SZLEV2) - 1;
    i[1] = (idx >> PGOBLIN_SZLEV3) & m;
    m = (1 << PGOBLIN_SZLEV3) - 1;
    i[2] = idx & m;
}

static void
/*****************************************************************************************************
 **                                                                                                 **/
decoxy(int x, int y, int *i) {                                                                     /**
 **                                                                                                 **
 *****************************************************************************************************/
    int    z[3];

    deco(x, i);
    deco(y, z);
    i[0] += (z[0] <<= PGOBLIN_SZLEV1);
    i[1] += (z[1] <<= PGOBLIN_SZLEV2);
    i[2] += (z[2] <<= PGOBLIN_SZLEV3);
}

static int
/*****************************************************************************************************
 **                                                                   **
 *****************************************************************************************************
 **                                                                                                 **/
formtouch(pgoblin_rst *rst, u_int64_t color, void *buf, int indx) {                                 /**
 **                                                                                                 **
 *****************************************************************************************************/
#   define blin_internal_flags (rst->style->flags & BLIN_MASK)
    int    form = rst->style->format;
    int    ex   = EX_OK; /*   */

    ifBLIN_QX6("+ %08X %"BLIN_X" f%d i%d", rst->style->flags, BLIN_I(buf), form, indx);
    switch(form) {
    case 0: {   u_int8_t *p = buf;

                p[indx] = (u_int8_t)color;
            }
          ; break;
    case 1: {   u_int16_t *p = buf;

                p[indx] = (u_int16_t)color;
            }
          ; break;
    case 2: {   u_int16_t *p = buf;

                p[indx] = (u_int16_t)color;
            }
          ; break;
    case 3: {   u_int32_t *p = buf;

                p[indx] = (u_int32_t)color;
            }
          ; break;
    case 4: {   u_int32_t *p = buf;

                p[indx] = (u_int32_t)color;
            }
          ; break;
    case 5: {   u_int64_t *p = buf;

                p[indx] = color;
            }
          ; break;
    case 6: {   u_int32_t *p = buf;

                p[indx] = (u_int32_t)color;
            }
          ; break;
    case 7: {   u_int64_t *p = buf;

                p[indx] = color;
            }
          ; break;
    case 8: {   u_int64_t *p = buf;

                if  (color & 1) {
                    p[indx >> 6] |= 1 << (indx & 0x03F);
                } else {
                    p[indx >> 6] &= ~(1 << (indx & 0x03F));
            }   }
          ; break;
    default:
        ifBLIN_QW0("format error");
        ex = -EX_DATAERR;
        errno = EINVAL;
    }
    ifBLIN_QX6("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static int
/*****************************************************************************************************
 **                                                                                                 **/
point(pgoblin_rst *rst, int x, int y, u_int64_t fg) {                                              /**
 **                                                                                                 **
 *****************************************************************************************************/
#   define blin_internal_flags (rst->flags & BLIN_MASK)
    int       form = rst->style->format;
    void   ***lev2;
    void    **lev3;
    int       ip[3];
    int       ex   = EX_OK; /*   */

    ifBLIN_QX3("+ %d %d %016"BLIN_O"X", x, y, fg);
    if  (  (x <= rst->style->marginxup  )
        && (x >= rst->style->marginxdown)
        && (y <= rst->style->marginyup  )
        && (y >= rst->style->marginydown)
        ) {
        if  (x < rst->style->fillxdown) rst->style->fillxdown = x;
        if  (x > rst->style->fillxup  ) rst->style->fillxup   = x;
        if  (y < rst->style->fillydown) rst->style->fillydown = y;
        if  (y > rst->style->fillyup  ) rst->style->fillyup   = y;
        decoxy(x, y, ip);
        ifBLIN_QX4(" %Xx%X [0]%02X [1]%02X [2]%04X", x, y, ip[0], ip[1], ip[2]);
        lev2 = &rst->style->grole[ip[0]];
        if  (!*lev2) {
            if  (!(*lev2 = calloc(1 << (PGOBLIN_SZLEV3 * 2), sizeof(void*)))) {
                ifBLIN_QW0("mem 2");
                ex = -EX_OSERR;
                goto out;
        }   }
        lev3 = &(*lev2)[ip[1]];
        if  (!*lev3) {
            if  (!(*lev3 = malloc(1 << (PGOBLIN_SZLEV3 * 2 + pgoblin_formlen[form] - 3)))) {
                ifBLIN_QW0("mem 3");
                ex = -EX_OSERR;
                goto out;
            }
            for (int i = 0; i < (1 << (PGOBLIN_SZLEV3 * 2)); ++i) {
                if  (!!(ex = formtouch(rst, rst->style->bg, *lev3, i))) {
                    ifBLIN_QW0("formtouch");
                    goto out;
        }   }   }
        ifBLIN_QX4("%"BLIN_X"[%d]", BLIN_I(*lev3), ip[2]);
        if  (!!(ex = formtouch(rst, fg, *lev3, ip[2]))) {
            ifBLIN_QW0("formtouch");
            goto out;
    }   }
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static int
/*****************************************************************************************************
 **                                                                                                 **/
line(pgoblin_rst *rst, int x1, int y1, int x2, int y2, u_int64_t fg) {                             /**
 **                                                                                                 **
 *****************************************************************************************************/
#   define blin_internal_flags (rst->flags & BLIN_MASK)
    int ex = EX_OK; /*   */

    ifBLIN_QX3("+ %d %d %d %d %016"BLIN_O"X", x1, y1, x2, y2, fg);
    if  (x1 > x2) {
        x1 ^= x2;
        x2 ^= x1;
        x1 ^= x2;
    }
    if  (y1 > y2) {
        y1 ^= y2;
        y2 ^= y1;
        y1 ^= y2;
    }
    if  (x1 == x2) {
        for (int i = y1; i <= y2; ++i) {
            if  (0 > (ex = point(rst, x1, i, fg))) {
                ifBLIN_QW0("point x %d", i);
                goto out;
        }   }
    } else if (y1 == y2) {
        for (int i = x1; i <= x2; ++i) {
            if  (0 > (ex = point(rst, i, y1, fg))) {
                ifBLIN_QW0("point y %d", i);
                goto out;
        }   }
    } else {
        ifBLIN_QX0("line (%d,%d),(%d,%d) %016"BLIN_O"X", x1, y1, x2, y2, fg);
    }
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static double
/*****************************************************************************************************
 **                                                                                                 **/
swaf(void *c) {                                                                                    /**
 **                                                                                                 **
 *****************************************************************************************************/
    union {
        double    f;
        u_int64_t t;
    } d;

    d.t = 0;
    for (int i = 0; i < 8; ++i) d.t = (d.t << 8) | ((u_char*)c)[i];
    return(d.f);
}

static int32_t
/*****************************************************************************************************
 **                                                                                                 **/
swai(void *c) {                                                                                    /**
 **                                                                                                 **
 *****************************************************************************************************/
    int32_t t;

    t = 0;
    for (int i = 0; i < 4; ++i) t = (t << 8) | ((u_char*)c)[i];
    return(t);
}

static int
/*****************************************************************************************************
 **                                                                                                 **/
scalexf(pgoblin_rst *rst, double f, int32_t *r) {                                                  /**
 **                                                                                                 **
 *****************************************************************************************************/
    int ex = EX_OK;
    double  d;

    d = round(f * rst->style->scalexup / rst->style->scalexdown);
    if  ((INT_MIN > d) || (INT_MAX < d)) {
        errno = EDOM;
        ex = -EX_DATAERR;
    } else {
        *r = (int32_t)lround(d);
    }
    return(ex);
}

static int
/*****************************************************************************************************
 **                                                                                                 **/
scaleyf(pgoblin_rst *rst, double f, int32_t *r) {                                                  /**
 **                                                                                                 **
 *****************************************************************************************************/
    int ex = EX_OK;
    double  d;

    d = round(f * rst->style->scaleyup / rst->style->scaleydown);
    if  ((INT_MIN > d) || (INT_MAX < d)) {
        errno = EDOM;
        ex = -EX_DATAERR;
    } else {
        *r = (int32_t)lround(d);
    }
    return(ex);
}

static int
/*****************************************************************************************************
 **                                                                                                 **/
scalex(pgoblin_rst *rst, void *sf, int32_t *r) {                                                   /**
 **                                                                                                 **
 *****************************************************************************************************/
    return(scalexf(rst, swaf(sf), r));
}

static int
/*****************************************************************************************************
 **                                                                                                 **/
scaley(pgoblin_rst *rst, void *sf, int32_t *r) {                                                   /**
 **                                                                                                 **
 *****************************************************************************************************/
    return(scaleyf(rst, swaf(sf), r));
}

static int
/*****************************************************************************************************
 **                                                                                                 **/
pgpoint(pgoblin_rst *rst, void *ppoint, u_int64_t fg) {                                            /**
 **                                                                                                 **
 *****************************************************************************************************/
#   define blin_internal_flags (rst->flags & BLIN_MASK)
    int32_t   tnp[2];
    char     *pnt = ppoint;
    ssize_t   ex  = EX_OK;

    if  (0 > (ex = scalex(rst, pnt, &tnp[0]))) goto out;
    pnt += 8;
    if  (0 > (ex = scaley(rst, pnt, &tnp[1]))) goto out;
    if  (0 > (ex = point(rst, tnp[0], tnp[1], fg))) {
        ifBLIN_QW0("point");
    }
out:
    ifBLIN_QX3("- %zd", ex);
    return((int)ex);
#   undef blin_internal_flags
}

static int
/*****************************************************************************************************
 **                                                                                                 **/
pglseg(pgoblin_rst *rst, void *plseg, u_int64_t fg) {                                              /**
 **                                                                                                 **
 *****************************************************************************************************/
#   define blin_internal_flags (rst->flags & BLIN_MASK)
    double    ends[4];
    int32_t   ges[4];
    char     *seg = plseg;
    ssize_t   ex  = 0;
    double    d;

    for (int i = 0; i < 4; ++i) {
        ends[i] = swaf(seg);
        seg += 8;
    }
    if  (fabs(ends[2] - ends[0]) > fabs(ends[3] - ends[1])) {
        if  (ends[0] > ends[2]) {
            seg = plseg;
            for (int i = 0; i < 4; ++i) {
                ends[i ^ 2] = swaf(seg);
                seg += 8;
        }   }
        if  (0 > (ex = scalexf(rst, ends[0], &ges[0]))) {
            ifBLIN_QW0("");
            goto out;
        }
        if  (0 > (ex = scalexf(rst, ends[2], &ges[2]))) {
            ifBLIN_QW0("");
            goto out;
        }
        for (int i = 0; i <= (ges[2] - ges[0]); ++i) {
            d = ((ends[1] * (ges[2] - ges[0] - i)) + (ends[3] * i)) / (ges[2] - ges[0]);
            if  (0 > (ex = scaleyf(rst, d, &ges[1]))) {
                ifBLIN_QW0("");
                goto out;
            }
            if  (0 > (ex = point(rst, ges[0] + i, ges[1], fg))) {
                ifBLIN_QW0("point in pglseg");
                goto out;
        }   }
    } else {
        if  (ends[1] > ends[3]) for (int i = 0; i < 4; ++i) {
            seg = plseg;
            for (int j = 0; j < 4; ++j) {
                ends[j ^ 2] = swaf(seg);
                seg += 8;
        }   }
        if  (0 > (ex = scaleyf(rst, ends[1], &ges[1]))) {
            ifBLIN_QW0("");
            goto out;
        }
        if  (0 > (ex = scaleyf(rst, ends[3], &ges[3]))) {
            ifBLIN_QW0("");
            goto out;
        }
        for (int i = 0; i <= (ges[3] - ges[1]); ++i) {
            d = ((ends[0] * (ges[3] - ges[1] - i)) + (ends[2] * i)) / (ges[3] - ges[1]);
            if  (0 > (ex = scalexf(rst, d, &ges[0]))) {
                ifBLIN_QW0("");
                goto out;
            }
            if  (0 > (ex = point(rst, ges[0], ges[1] + i, fg))) {
                ifBLIN_QW0("point in pglseg");
                goto out;
    }   }   }
out:
    ifBLIN_QX3("- %zd", ex);
    return((int)ex);
#   undef blin_internal_flags
}

static int
/*****************************************************************************************************
 **                                                                                                 **/
pgbox(pgoblin_rst *rst, void *pbox, u_int64_t fg) {                                                /**
 **                                                                                                 **
 *****************************************************************************************************/
#   define blin_internal_flags (rst->flags & BLIN_MASK)
    int32_t   xob[4];
    char     *box = pbox;
    ssize_t   ex  = EX_OK;

    for (int i = 0; i < 4;) {
        if  (!!(ex = scalex(rst, box, &xob[i++]))) goto out;
        box += 8;
        if  (!!(ex = scaley(rst, box, &xob[i++]))) goto out;
        box += 8;
    }
    if  (0 > (ex = line(rst, xob[0], xob[1], xob[2], xob[1], fg))) {
        ifBLIN_QW0("line 0");
        goto out;
    }
    if  (0 > (ex = line(rst, xob[0], xob[3], xob[2], xob[3], fg))) {
        ifBLIN_QW0("line 1");
        goto out;
    }
    if  (0 > (ex = line(rst, xob[0], xob[1], xob[0], xob[3], fg))) {
        ifBLIN_QW0("line 2");
        goto out;
    }
    if  (0 > (ex = line(rst, xob[2], xob[1], xob[2], xob[3], fg))) {
        ifBLIN_QW0("line 3");
        goto out;
    }
out:
    ifBLIN_QX3("- %zd", ex);
    return((int)ex);
#   undef blin_internal_flags
}

static int
/*****************************************************************************************************
 **                                                                                                 **/
pgpath(pgoblin_rst *rst, void *ppath, u_int64_t fg, u_int64_t fg1) {                               /**
 **                                                                                                 **
 *****************************************************************************************************/
#   define blin_internal_flags (rst->flags & BLIN_MASK)
    int       npts;
    u_char   *path = ppath;
    double    seg[4];
    ssize_t   ex   = EX_OK;
    int       cl;

    cl = *(path++);
    npts = swai(path);
    path += 4;
    if  (2 > npts) goto out;
    bcopy(path, &seg[2], 2 * sizeof(double));
    for (int i = 1; i < npts; ++i) {
        if  (0 > (ex = pglseg(rst, path, fg))) {
            ifBLIN_QW0("path line N%d", i);
            goto out;
        }
        path += 16;
    }
    bcopy(path, seg, 2 * sizeof(double));
    if  (!!cl && (0 > (ex = pglseg(rst, seg, fg1)))) {
        ifBLIN_QW0("path line last");
        goto out;
    }
out:
    ifBLIN_QX3("- %zd", ex);
    return((int)ex);
#   undef blin_internal_flags
}

static int
/*****************************************************************************************************
 **                                                                                                 **/
pgpoly(pgoblin_rst *rst, void *ppoly, u_int64_t fg, u_int64_t fg1) {                               /**
 **                                                                                                 **
 *****************************************************************************************************/
#   define blin_internal_flags (rst->flags & BLIN_MASK)
    int       npts;
    u_char   *poly = ppoly;
    double    pol[4];
    ssize_t   ex   = EX_OK;

    npts = swai(poly);
    poly += 4;
    if  (2 > npts) goto out;
    bcopy(poly, &pol[2], 2 * sizeof(double));
    for (int i = 1; i < npts; ++i) {
        if  (0 > (ex = pglseg(rst, poly, fg))) {
            ifBLIN_QW0("poly line N%d", i);
            goto out;
        }
        poly += 16;
    }
    bcopy(poly, pol, 2 * sizeof(double));
    if  (0 > (ex = pglseg(rst, pol, fg1))) {
        ifBLIN_QW0("poly line last");
        goto out;
    }
out:
    ifBLIN_QX3("- %zd", ex);
    return((int)ex);
#   undef blin_internal_flags
}

static int
/*****************************************************************************************************
 ****                                                                                             ****/
findnm(pgoblin_rst *rst, confar *fgs, const char *colname, u_int64_t *fg) {                      /****
 ****                                                                                             ****
 *****************************************************************************************************/
#   define blin_internal_flags (rst->flags & BLIN_MASK)
    int r = 0;

    (void)rst;
    (void)blin_internal_flags;
    for (int i = 0; i < fgs->nmsz; ++i) {
        if  (!strcmp(fgs->namec[i].name, colname)) {
            *fg = fgs->namec[i].fg;
            r = 1;
            break;
    }   }
    return(r);
#   undef blin_internal_flags
}

static int
/*****************************************************************************************************
 ****                                                                                             ****/
dofg(pgoblin_rst *rst, u_int8_t type, ssize_t column, const char *colname, u_int64_t *fg) {      /****
 ****                                                                                             ****
 *****************************************************************************************************/
#   define blin_internal_flags (rst->flags & BLIN_MASK)
    int r = 0;

    (void)blin_internal_flags;
    if  (!!rst->style->fg[type]) {
        if  (column < rst->style->sz[type]) {
            if  (!!(r = findnm(rst, &rst->style->fg[type][column + 1], colname, fg))) goto out;
        }
        if  (!!(r = findnm(rst, &rst->style->fg[type][0], colname, fg))) goto out;
        if  ((column < rst->style->sz[type]) && (rst->style->fg[type][column + 1].def)) {
            *fg = rst->style->fg[type][column + 1].fg;
            r = 1;
            goto out;
        }
        if  (rst->style->fg[type][0].def) {
            *fg = rst->style->fg[type][0].fg;
            r = 1;
            goto out;
    }   }
out:
    return(r);
#   undef blin_internal_flags
}

static ssize_t
/*****************************************************************************************************
 *****************************************************************************************************
 ****                                                                                             ****/
table(pgoblin_exenv *exenv, pgoblin_nr *rn, ssize_t low, ssize_t hig) {                          /****
 ****                                                                                             ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (exenv->options->flags & BLIN_MASK)
    const char  *typename;
    int          form;
    ssize_t      ntyp;
    ssize_t      ncol;
    void       **lev2;
    void        *lev3;
    pgoblin_rio *rou  = NULL;
    pgoblin_rio *rin;
    pgoblin_rst *rst;
    FILE        *fi   = NULL;
    ssize_t      ex   = -EX_DATAERR;
    ssize_t      e;

    ifBLIN_QX2("+ [%"BLIN_D",%"BLIN_D"]", low, hig);
    GET_STY(rst, exenv->options, rn[PGO_CSTY]);
    GET_RIO(rou, exenv->options, rn[PGO_COUT]);
    GET_RIO(rin, exenv->options, rn[PGO_CIN]);
    ex = 0;
    ifBLIN_QX3(" %"BLIN_X" %"BLIN_X" %"BLIN_X, BLIN_I(rou), BLIN_I(rst), BLIN_I(rin));
    form = rst->style->format;
    if  (!!strncmp(rst->style->id, PGOBLIN_STYLE_ID, strlen(PGOBLIN_STYLE_ID))) {
        ifBLIN_QX0("Illegal style %s @STY %c", rst->style->id, pgoblin_regn[rn[PGO_CSTY]]);
        ERROUT(-EX_DATAERR, EINVAL);
    }
    if  (0 >= (ntyp = (ssize_t)pgoblin_db_resinfo(rin, PGOBLIN_Ntuples))) goto out;
    if  (0 > (ncol = (ssize_t)pgoblin_db_resinfo(rin, PGOBLIN_Nfields))) {
        ifBLIN_QW0("resinfo Nfields");
        ex = -EX_SOFTWARE;
        goto out;
    }
    for (ssize_t k = 0; k < ncol; ++k) {
        const char *colname;
        void       *geobj;
        u_int8_t    type;
        u_int64_t   fg1;
        u_int64_t   fg;
/*
 * :
 *   
 *  
 *      
 * 
 * ---  
 * --- 
 * ---     
 * ---
 */
        if  (!(typename = pgoblin_db_subinfo(rin, PGOBLIN_TypeNm, k))) continue;
        type = (u_int8_t)babolo_testword(&gtype, typename);
        colname = pgoblin_db_subinfo(rin, PGOBLIN_ColName, k);
        for (ssize_t i = 0; i < ntyp; ++i) {
            if  (  !pgoblin_db_valinfo(rin, PGOBLIN_IsNull, i, k)
                && !!(geobj = pgoblin_db_getvalue(rin, i, k))
                ) {
                if  (!!dofg(rst, type, k, colname, &fg)) goto found;
                if  (!!dofg(rst, 0, k, colname, &fg)) goto found;
                fg = rst->style->fg0;
found:
                switch(type) {
                case PGOBLIN_GEO_POINT:
                    if  (0 > (ex = pgpoint(rst, geobj, fg))) {
                        ifBLIN_QW0("point");
                        goto out;
                    }
                    break;
                case PGOBLIN_GEO_LSEG:
                    if  (0 > (ex = pglseg(rst, geobj, fg))) {
                        ifBLIN_QW0("lseg");
                        goto out;
                    }
                    break;
                case PGOBLIN_GEO_BOX:
                    if  (0 > (ex = pgbox(rst, geobj, fg))) {
                        ifBLIN_QW0("box");
                        goto out;
                    }
                    break;
                case PGOBLIN_GEO_PATH:
                    fg1 = fg;
                    dofg(rst, PGOBLIN_GEO_PATHP, k, colname, &fg1);
                    if  (0 > (ex = pgpath(rst, geobj, fg, fg1))) {
                        ifBLIN_QW0("path");
                        goto out;
                    }
                    break;
                case PGOBLIN_GEO_POLYGON:
                    fg1 = fg;
                    dofg(rst, PGOBLIN_GEO_POLYP, k, colname, &fg1);
                    if  (0 > (ex = pgpoly(rst, geobj, fg, fg1))) {
                        ifBLIN_QW0("poly");
                        goto out;
                    }
                    break;
                default:
                    ifBLIN_QX0("Unknown geometric object");
                    ex = -EX_SOFTWARE;
                    goto out;
                }
                if  (0 < ex) {
                    ex = 0;
                    continue;
    }   }   }   }
    if  (  (rst->style->fillxdown <= rst->style->fillxup)
        && (rst->style->fillydown <= rst->style->fillyup)
        ) {
        int    sz;

        sz = 1 << (PGOBLIN_SZLEV1 + PGOBLIN_SZLEV2 + PGOBLIN_SZLEV3 - 1);
        if  (rst->style->marginxup   == sz - 1) rst->style->marginxup   = rst->style->fillxup;
        if  (rst->style->marginyup   == sz - 1) rst->style->marginyup   = rst->style->fillyup;
        if  (rst->style->marginxdown == -sz   ) rst->style->marginxdown = rst->style->fillxdown;
        if  (rst->style->marginydown == -sz   ) rst->style->marginydown = rst->style->fillydown;
        if  (!(fi = pgoblin_io_funopen(rou))) {
            ifBLIN_QW0("funopen on %d", rn[PGO_COUT]);
            ex = -EX_IOERR;
            goto out;
        }
        e = fprintf( fi
                   , "P7\nWIDTH %u\nHEIGHT %u\n"
                   , rst->style->marginxup - rst->style->marginxdown + 1
                   , rst->style->marginyup - rst->style->marginydown + 1
                   );
        if  (0 > e) {
            ifBLIN_QW0("write error");
            ex = -EX_IOERR;
            goto out;
        } else {
            ex += e;
        }
        switch(form) {
        case 0: e = fprintf(fi, "DEPTH 1\nMAXVAL 255\nTUPLTYPE GRAYSCALE\nENDHDR\n"        ); break;
        case 1: e = fprintf(fi, "DEPTH 1\nMAXVAL 65535\nTUPLTYPE GRAYSCALE\nENDHDR\n"      ); break;
        case 2: e = fprintf(fi, "DEPTH 2\nMAXVAL 255\nTUPLTYPE GRAYSCALE_ALPHA\nENDHDR\n"  ); break;
        case 3: e = fprintf(fi, "DEPTH 2\nMAXVAL 65535\nTUPLTYPE GRAYSCALE_ALPHA\nENDHDR\n"); break;
        case 4: e = fprintf(fi, "DEPTH 3\nMAXVAL 255\nTUPLTYPE RGB\nENDHDR\n"              ); break;
        case 5: e = fprintf(fi, "DEPTH 3\nMAXVAL 65535\nTUPLTYPE RGB\nENDHDR\n"            ); break;
        case 6: e = fprintf(fi, "DEPTH 4\nMAXVAL 255\nTUPLTYPE RGB_ALPHA\nENDHDR\n"        ); break;
        case 7: e = fprintf(fi, "DEPTH 4\nMAXVAL 65535\nTUPLTYPE RGB_ALPHA\nENDHDR\n"      ); break;
        case 8: e = fprintf(fi, "DEPTH 1\nMAXVAL 1\nTUPLTYPE BLACKANDWHITE\nENDHDR\n"      ); break;
        default:
            ifBLIN_QW0("format error");
            ex = -EX_DATAERR;
            goto out;
        }
        if  (0 > e) {
            ifBLIN_QW0("write error");
            ex = -EX_IOERR;
            goto out;
        } else {
            ex += e;
        }
        pgoblin_io_funclose(rou);
        for (int y = rst->style->marginydown; y <= rst->style->marginyup; ++y) {
            int    xi[3];
            int    yi[3];

            deco(y, yi);
            yi[0] <<= PGOBLIN_SZLEV1;
            yi[1] <<= PGOBLIN_SZLEV2;
            yi[2] <<= PGOBLIN_SZLEV3;
            for (int x = rst->style->marginxdown; x <= rst->style->marginxup; ++x) {
                deco(x, xi);
                lev2 = rst->style->grole[yi[0] + xi[0]];
                lev3 = NULL;
                if  (!!lev2) lev3 = lev2[yi[1] + xi[1]];
                if  (!!lev3) {
                    int idx = yi[2] + xi[2];

                    switch(form) {
                    case 0: {   u_int8_t *p = lev3;

                                e = pgoblin_io_write(rou, &p[idx], 1);
                            }
                          ; break;
                    case 1: {   u_int16_t *p = lev3;

                                e = pgoblin_io_write(rou, &p[idx], 2);
                            }
                          ; break;
                    case 2: {   u_int16_t *p = lev3;

                                e = pgoblin_io_write(rou, &p[idx], 2);
                            }
                          ; break;
                    case 3: {   u_int32_t *p = lev3;

                                e = pgoblin_io_write(rou, &p[idx], 4);
                            }
                          ; break;
                    case 4: {   u_int32_t *p = lev3;

                                e = pgoblin_io_write(rou, &p[idx], 3);
                            }
                          ; break;
                    case 5: {   u_int64_t *p = lev3;

                                e = pgoblin_io_write(rou, &p[idx], 6);
                            }
                          ; break;
                    case 6: {   u_int32_t *p = lev3;

                                e = pgoblin_io_write(rou, &p[idx], 4);
                            }
                          ; break;
                    case 7: {   u_int64_t *p = lev3;

                                e = pgoblin_io_write(rou, &p[idx], 8);
                            }
                          ; break;
                    case 8: {   ifBLIN_QX0("Not now");
                            }
                          ; break;
                    default:
                        ifBLIN_QW0("format error");
                        ex = -EX_DATAERR;
                        goto out;
                    }
                } else {
                    e = pgoblin_io_write(rou, &rst->style->bg, pgoblin_formbyt[form] / 2);
                }
                if  (0 > e) {
                    ifBLIN_QW0("write error");
                    ex = -EX_IOERR;
                    goto out;
                } else {
                    ex += e;
    }   }   }   }
    ifBLIN_QX2( "%08X %d..%d x %d..%d = %d x %d"
              , rst->style->flags
              , rst->style->marginxdown
              , rst->style->marginxup
              , rst->style->marginydown
              , rst->style->marginyup
              , rst->style->marginxup - rst->style->marginxdown + 1
              , rst->style->marginyup - rst->style->marginydown + 1
              );
    ifBLIN_QX2( "w%u b%016"BLIN_O"X %f/%f x %f/%f"
              , rst->style->wideline
              , rst->style->bg
              , rst->style->scalexup
              , rst->style->scalexdown
              , rst->style->scaleyup
              , rst->style->scaleydown
              );
    ifBLIN_QX2( "f%016"BLIN_O"X f%016"BLIN_O"X f%016"BLIN_O"X f%016"BLIN_O"X"
                " f%016"BLIN_O"X f%016"BLIN_O"X f%016"BLIN_O"X f%016"BLIN_O"X"
              , rst->style->fg[0]->fg
              , rst->style->fg[1]->fg
              , rst->style->fg[2]->fg
              , rst->style->fg[3]->fg
              , rst->style->fg[4]->fg
              , rst->style->fg[5]->fg
              , rst->style->fg[6]->fg
              , rst->style->fg[7]->fg
              );
out:
    ifBLIN_QX2("- %"BLIN_D, ex);
    return(ex);
#   undef blin_internal_flags
}

pgoblin_styles PGOBLIN_MODULE =
{ 0
, "5pGoblin-" VERS
, "pgeo\0" VERS
, 0
, init
, fini
, parser
, dump
, clear
, NULL /* get    */
, table
, NULL /* load   */
, NULL /* store  */
, NULL /* dummy0 */
, NULL /* dummy1 */
};
